#version 410

#ifdef VERTEX_SHADER

in vec3 in_position;
in vec2 in_texcoord;

in int in_codepoint;
in vec3 in_glyphPosition;
in vec4 in_rotation;
in vec2 in_glyphDimensions;
in vec4 in_color;
in vec2 in_scale;
in vec2 in_anchor;

in float in_thickness;
in float in_outline;
in vec4 in_outlineColor;

in vec4 in_shadowColor;
in float in_shadowThickness;
in float in_shadowSoftness;
in vec2 in_shadowOffset;

out vec2 texcoord;
out vec4 color;
out vec2 uv;
out vec4 texcoordFull;

uniform sampler1D texcoordsMap;

uniform mat4 model;

#include "shaders/lib/camera.glsl"

void main() {
    float xSquare = in_rotation.x * in_rotation.x;
    float ySquare = in_rotation.y * in_rotation.y;
    float zSquare = in_rotation.z * in_rotation.z;

    vec3 rot_col0 = vec3(1 - (2 * ySquare) - (2 * zSquare),
                         (2 * in_rotation.x * in_rotation.y) - (2 * in_rotation.w * in_rotation.z),
                         (2 * in_rotation.x * in_rotation.z) + (2 * in_rotation.w * in_rotation.y));

    vec3 rot_col1 = vec3((2 * in_rotation.x * in_rotation.y) + (2 * in_rotation.w * in_rotation.z),
                         1 - (2 * xSquare) - (2 * zSquare),
                         (2 * in_rotation.y * in_rotation.z) - (2 * in_rotation.w * in_rotation.x));

    vec3 rot_col2 = vec3((2 * in_rotation.x * in_rotation.z) - (2 * in_rotation.w * in_rotation.y),
                         (2 * in_rotation.y * in_rotation.z) + (2 * in_rotation.w * in_rotation.x),
                         1 - (2 * xSquare) - (2 * ySquare));

    vec3 modelSpaceAnchorPosition = in_glyphPosition + vec3(in_glyphDimensions * in_anchor, 0);

    mat3 rot_scale = mat3(in_scale.x * in_glyphDimensions.x * rot_col0,
                          in_scale.y * in_glyphDimensions.y * rot_col1,
                          rot_col2);
    vec3 modelPosition = rot_scale * (vec3(in_position.xz - in_anchor, 0));
    modelPosition += modelSpaceAnchorPosition;

    gl_Position = viewProjection * (model * vec4(modelPosition, 1));

    vec4 texcoords = texelFetch(texcoordsMap, in_codepoint, 0);
    vec2 texcoord0 = texcoords.xy;
    vec2 texcoord1 = texcoords.zw;
    texcoord = mix(texcoord0, texcoord1, in_texcoord);

    uv = in_texcoord;
    texcoordFull = texcoords;

    color = in_color;
}

#endif

#ifdef FRAGMENT_SHADER

in vec4 texcoordFull;
in vec2 texcoord;
in vec4 color;
in vec2 uv;

out vec4 fragColor;

uniform sampler2D atlas;
uniform float msdfRange;

float median(float r, float g, float b) {
    return max(min(r, g), min(max(r, g), b));
}

float LinearStep(float a, float b, float x) {
    return (x - a) / (b - a);
}

void main() {
    vec4 s = texture(atlas, texcoord);

    float thickness = 0.0; // -1 - 1
    float border = 0.0; // 0-0.25

    float shadowThickness = 0.0;

    float shadowOpacity = color.a; // 0-1
    float shadowSoftness = 0.0; // 0-1

    float sigDist = 2.0 * median(s.r, s.g, s.b) - 1.0 + thickness;
    
    vec4 tmpColor = color;

    float opacity = 0.0;

    vec2 atlasSize = textureSize(atlas, 0);
    

    float pxRange = atlasSize.x / msdfRange;
#if 1
    float pxSize = min(0.5 / pxRange * (fwidth(uv.x) * atlasSize.x + fwidth(uv.y) * atlasSize.y), 0.25);

    float inside = LinearStep(-border - pxSize, -border + pxSize, sigDist);
    float outside = 1.0;
    if (border > 0) {
        outside = LinearStep(border - pxSize, border + pxSize, sigDist);
    }

    tmpColor.rgb = mix(vec3(0, 0, 0), tmpColor.rgb, outside);
    opacity = clamp(inside, 0, 1);

    // @NOTE: we have to clamp the value of inside and use that for opacity, otherwise
    // the shadow calculations break
    tmpColor.a *= opacity;

#if 1
    vec2 shadowOffset = vec2(-0.02, -0.02); // very small range before we start grabbing other glyphs
    vec2 texcoord0 = texcoordFull.xy;
    vec2 texcoord1 = texcoordFull.zw;
    vec2 shadowTexcoord = mix(texcoord0, texcoord1, uv - shadowOffset);
    s = texture(atlas, shadowTexcoord);

    sigDist = 2.0 * median(s.r, s.g, s.b) - 1.0 + border + shadowThickness;
    float shadow = shadowOpacity * LinearStep(-shadowSoftness - pxSize, shadowSoftness + pxSize, sigDist);
    
    tmpColor.rgb = mix(vec3(0, 0, 0), tmpColor.rgb, tmpColor.a);
    tmpColor.a = ((shadow - shadow * tmpColor.a + tmpColor.a));
#endif
    
    fragColor = tmpColor;

    
#endif

    // @NOTE: this guys method seems to have less aliasing artifacts than Chlumsky, but its a bit fidly
#if 0
    // https://github.com/Chlumsky/msdfgen/issues/22
    vec2 sz = textureSize(atlas, 0);
    // sz.x *= msdfRange;
    // sz.y *= msdfRange;
    float dx = dFdx( texcoord.x ) * sz.x;
    float dy = dFdy( texcoord.y ) * sz.y;

    // @GACK @BUG: this mult factor only works when our textures are 1024x1024, if they're bigger it needs to
    // be bigger too
    float toPixels = 8.0 * inversesqrt( dx * dx + dy * dy );
    opacity = clamp( sigDist * toPixels + 0.5, 0.0, 1.0 );
    fragColor = color;
    fragColor.a *= opacity;
    
#endif
    

    
#define RENDER_MSDF 0
    
#if RENDER_MSDF
    fragColor = s;
#endif
}



#endif

